/* MESA private include file
   (c) 1998 Lionel ULMER
   Copyright 2000-2001 TransGaming Technologies Inc.

   This file contains all structures that are not exported
   through d3d.h and all common macros. */

#ifndef __GRAPHICS_WINE_MESA_PRIVATE_H
#define __GRAPHICS_WINE_MESA_PRIVATE_H

#include "wine/wine_gl.h"

#ifdef HAVE_OSMESA
#include <GL/osmesa.h>
#endif

#include "d3d_private.h"

#include "x11drv.h"

/*****************************************************************************
 * IDirect3DLight MESA private structure
 */
typedef struct mesa_d3dl_private
{
    GLenum              light_num;
} mesa_d3dl_private;

/*****************************************************************************
 * IDirect3DTexture2 MESA private structure
 */
typedef struct mesa_d3dt_private
{
    GLuint                   tex_name;
} mesa_d3dt_private;

/*****************************************************************************
 * IDirect3DViewport2 implementation structure
 */
typedef struct mesa_d3dv_private
{
    GLenum                      nextlight;
} mesa_d3dv_private;

/* Matrix copy WITH transposition */
#define conv_mat2(mat,gl_mat)			\
{						\
    TRACE("%f %f %f %f\n", (mat)->_11, (mat)->_12, (mat)->_13, (mat)->_14); \
    TRACE("%f %f %f %f\n", (mat)->_21, (mat)->_22, (mat)->_23, (mat)->_24); \
    TRACE("%f %f %f %f\n", (mat)->_31, (mat)->_32, (mat)->_33, (mat)->_34); \
    TRACE("%f %f %f %f\n", (mat)->_41, (mat)->_42, (mat)->_43, (mat)->_44); \
    (gl_mat)[ 0] = (mat)->_11;			\
    (gl_mat)[ 1] = (mat)->_21;			\
    (gl_mat)[ 2] = (mat)->_31;			\
    (gl_mat)[ 3] = (mat)->_41;			\
    (gl_mat)[ 4] = (mat)->_12;			\
    (gl_mat)[ 5] = (mat)->_22;			\
    (gl_mat)[ 6] = (mat)->_32;			\
    (gl_mat)[ 7] = (mat)->_42;			\
    (gl_mat)[ 8] = (mat)->_13;			\
    (gl_mat)[ 9] = (mat)->_23;			\
    (gl_mat)[10] = (mat)->_33;			\
    (gl_mat)[11] = (mat)->_43;			\
    (gl_mat)[12] = (mat)->_14;			\
    (gl_mat)[13] = (mat)->_24;			\
    (gl_mat)[14] = (mat)->_34;			\
    (gl_mat)[15] = (mat)->_44;			\
};

/* Matrix copy WITHOUT transposition */
#define conv_mat(mat,gl_mat)			\
{                                               \
    TRACE("%f %f %f %f\n", (mat)->_11, (mat)->_12, (mat)->_13, (mat)->_14); \
    TRACE("%f %f %f %f\n", (mat)->_21, (mat)->_22, (mat)->_23, (mat)->_24); \
    TRACE("%f %f %f %f\n", (mat)->_31, (mat)->_32, (mat)->_33, (mat)->_34); \
    TRACE("%f %f %f %f\n", (mat)->_41, (mat)->_42, (mat)->_43, (mat)->_44); \
    memcpy(gl_mat, (mat), 16 * sizeof(float));      \
};

typedef struct render_state {
  /* This is used for the device mode */
  GLenum src, dst;
  /* This is used for textures */
  GLenum mag, min;
} RenderState;

#define NUM_TEXTURE_STAGES 8

struct mesa_texture_stage
{
    /* SetTexture */
    IDirectDrawSurface7 *texture;

    /* SetTextureStageState */
    BOOL enabled;
    D3DTEXTUREOP colorop;
    DWORD colorarg1;
    DWORD colorarg2;
    D3DTEXTUREOP alphaop;
    DWORD alphaarg1;
    DWORD alphaarg2;
    DWORD coord_idx;
    D3DTEXTUREMIPFILTER mipfilter;
    D3DTEXTUREMINFILTER minfilter;
    D3DTEXTUREMINFILTER magfilter;
    D3DTEXTUREADDRESS u_addr, v_addr;
    float lodbias;
    DWORD maxmiplevel;
};

struct mesa_light
{
    BOOL enabled;
    DWORD view_generation;
    D3DLIGHT7 l;
    float pos[4];
};

typedef enum
{
    ARB_multitexture,
    ARB_texture_compression,
    ARB_texture_cube_map,
    ARB_texture_env_add,
    EXT_blend_func_separate,
    EXT_fog_coord,
    EXT_secondary_color,
    EXT_stencil_wrap,
    EXT_texture_compression_s3tc,
    EXT_texture_env_add,
    EXT_texture_env_combine,
    EXT_texture_lod_bias,
    EXT_vertex_weighting,
    NV_blend_square,
    S3_s3tc,
    EXT_clip_volume_hint,
    ARB_texture_env_combine,
    EXT_texture_env_dot3,
    NUM_GL_EXTENSIONS
} gl_extensions;

typedef struct mesa_d3dd_private {
    GLXContext glx_ctx;
    GLXDrawable drawable;
#ifdef GLX_VERSION_1_3
    GLXFBConfig *fbconfig;
#endif
    XVisualInfo *vis;
    BOOL using_primary;
#ifdef HAVE_OSMESA
    OSMesaContext osm_ctx;
    unsigned char *surface;
#endif

    /* The current render state */
    RenderState rs;

    /* The last type of vertex drawn */
    D3DVERTEXTYPE vt;

    GLint drawable_width, drawable_height;

    LPVOID color_array;

    D3DMATRIX world_mat;
    D3DMATRIX view_mat;
    D3DMATRIX proj_mat;

    BOOL color_vertex;
    BOOL specular;
    D3DMATERIALCOLORSOURCE diffuse_source;
    D3DMATERIALCOLORSOURCE specular_source;
    D3DMATERIALCOLORSOURCE ambient_source;
    D3DMATERIALCOLORSOURCE emissive_source;
    D3DMATERIAL7 material;

    DWORD num_lights;
    struct mesa_light* lights;

    BOOL transform_changed;
    BOOL view_changed;
    DWORD view_generation;

    BOOL fog_enable;
    BOOL clip_enable;
    D3DFOGMODE vertex_fog_mode, table_fog_mode;

    GLclampf alpha_ref;
    GLenum alpha_func;

    GLenum src_blend, dst_blend;

    GLenum stencil_fail_op, stencil_zfail_op, stencil_pass_op;
    GLenum stencil_func;
    GLint stencil_ref;
    GLuint stencil_read_mask;

    HRESULT (*texture_combiner)(struct mesa_d3dd_private*, DWORD,
				D3DTEXTURESTAGESTATETYPE, DWORD);
    GLint max_gl_textures;

    PFNGLACTIVETEXTUREARBPROC glActiveTextureARB;
    PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureARB;
    PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2fARB;
    PFNGLSECONDARYCOLOR3UBEXTPROC glSecondaryColor3ubEXT;
    PFNGLSECONDARYCOLORPOINTEREXTPROC glSecondaryColorPointerEXT;

    struct mesa_texture_stage texstage[NUM_TEXTURE_STAGES];

    BYTE extension_present[NUM_GL_EXTENSIONS];
} mesa_d3dd_private;

#define _dump_colorvalue(s,v)               \
    TRACE(" " s " : %f %f %f %f\n",           \
	    (v).u1.r, (v).u2.g, (v).u3.b, (v).u4.a);

/* Common functions defined in d3dcommon.c */
void set_render_state(D3DRENDERSTATETYPE dwRenderStateType,
		      DWORD dwRenderState, RenderState *rs) ;

/* All non-static functions 'exported' by various sub-objects */
extern LPDIRECT3DTEXTURE2 d3dtexture2_create(IDirectDrawSurfaceImpl* surf);
extern LPDIRECT3DTEXTURE d3dtexture_create(IDirectDrawSurfaceImpl* surf);

extern LPDIRECT3DLIGHT d3dlight_create_dx3(IDirect3DImpl* d3d1);
extern LPDIRECT3DLIGHT d3dlight_create(IDirect3D2Impl* d3d2);

extern LPDIRECT3DEXECUTEBUFFER d3dexecutebuffer_create(IDirect3DDeviceImpl* d3ddev, LPD3DEXECUTEBUFFERDESC lpDesc);

extern LPDIRECT3DMATERIAL d3dmaterial_create(IDirect3DImpl* d3d1);
extern LPDIRECT3DMATERIAL2 d3dmaterial2_create(IDirect3D2Impl* d3d2);

extern LPDIRECT3DVIEWPORT d3dviewport_create(IDirect3DImpl* d3d1);
extern LPDIRECT3DVIEWPORT2 d3dviewport2_create(IDirect3D2Impl* d3d2);

extern int d3denum_OpenGL(LPD3DENUMDEVICESCALLBACK cb, LPVOID context);
extern int d3d7enum_OpenGL(LPD3DENUMDEVICESCALLBACK7 cb, LPVOID context);
extern int d3dinst_OpenGL(REFCLSID rguid, IDirectDrawSurfaceImpl* surface, IDirect3DDeviceImpl** device, IDirect3DImpl* d3d, LPVOID vtable);
extern int d3d7inst_OpenGL(REFCLSID rguid, IDirectDrawSurfaceImpl* surface, IDirect3DDeviceImpl** device, IDirect3DImpl* d3d, LPVOID vtable);

extern LPDIRECT3DTEXTURE2 mesa_d3dtexture2_create(IDirectDrawSurfaceImpl*);
extern LPDIRECT3DTEXTURE mesa_d3dtexture_create(IDirectDrawSurfaceImpl*);

extern HRESULT WINAPI MESA_IDirect3DDevice7Impl_Clear(LPDIRECT3DDEVICE7 iface,
						      DWORD dwCount,LPD3DRECT lpRects,DWORD dwFlags,
						      D3DCOLOR dwColor,D3DVALUE dvZ,DWORD dwStencil);

static const GUID WINE_UNUSED IID_D3DDEVICE2_OpenGL = {
  0x39a0da38,
  0x7e57,
  0x11d2,
  { 0x8b,0x7c,0x0e,0x4e,0xd8,0x3c,0x2b,0x3c }
};

static const GUID WINE_UNUSED IID_D3DDEVICE_OpenGL = {
  0x31416d44,
  0x86ae,
  0x11d2,
  { 0x82,0x2d,0xa8,0xd5,0x31,0x87,0xca,0xfa }
};

/* This structure contains all the function pointers to OpenGL extensions
   that are used by Wine */
typedef struct {
  void (*ptr_ColorTableEXT) (GLenum target, GLenum internalformat,
			     GLsizei width, GLenum format, GLenum type, const GLvoid *table);
} Mesa_DeviceCapabilities;

extern ICOM_VTABLE(IDirect3D) mesa_d3dvt;
extern ICOM_VTABLE(IDirect3D2) mesa_d3d2vt;
extern ICOM_VTABLE(IDirect3D3) mesa_d3d3vt;
extern ICOM_VTABLE(IDirect3D7) mesa_d3d7vt;

extern ICOM_VTABLE(IDirect3DDevice2) OpenGL_d3dd2vt;
extern ICOM_VTABLE(IDirect3DDevice7) OpenGL_d3dd7vt;

#endif /* __GRAPHICS_WINE_MESA_PRIVATE_H */
