/*
 * Server-side change notification management
 *
 * Copyright (C) 1998 Alexandre Julliard
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "winnt.h"

#include "handle.h"
#include "thread.h"
#include "request.h"

struct change
{
    struct object  obj;      /* object header */
    int            subtree;  /* watch all the subtree */
    int            filter;   /* notification filter */
};

static void change_dump( struct object *obj, int verbose );
static int change_signaled( struct object *obj, struct wait_queue_entry *entry );

static const struct object_ops change_ops =
{
    sizeof(struct change),    /* size */
    change_dump,              /* dump */
    add_queue,                /* add_queue */
    remove_queue,             /* remove_queue */
    change_signaled,          /* signaled */
    no_satisfied,             /* satisfied */
    NULL,                     /* get_poll_events */
    NULL,                     /* poll_event */
    no_get_fd,                /* get_fd */
#if defined( CONTEXTUAL_FD )
    no_get_ctx_fd,            /* get_ctx_fd */
#endif
    no_flush,                 /* flush */
    no_get_file_info,         /* get_file_info */
    NULL,                     /* queue_async */
    no_destroy                /* destroy */
};
const struct object_ops* get_this_process_change_ops(void) { return &change_ops; }

static struct change *create_change_notification( int subtree, int filter )
{
    struct change *change;
    if ((change = alloc_object( change_ops_tag, -1 )))
    {
        change->subtree = subtree;
        change->filter  = filter;
    }
    return change;
}

static void change_dump( struct object *obj, int verbose )
{
    struct change *change = (struct change *)obj;
    assert( obj->ops == change_ops_tag );
    fprintf( stderr, "Change notification sub=%d filter=%08x\n",
             change->subtree, change->filter );
}

static int change_signaled( struct object *obj, struct wait_queue_entry *entry )
{
/*    struct change *change = (struct change *)obj;*/
    assert( obj->ops == change_ops_tag );
    return 0;  /* never signaled for now */
}

/* create a change notification */
DECL_HANDLER(create_change_notification)
{
    struct change *change;

    reply->handle = 0;
    if ((change = create_change_notification( req->subtree, req->filter )))
    {
        reply->handle = alloc_handle( current->process, change,
                                      STANDARD_RIGHTS_REQUIRED|SYNCHRONIZE, 0 );
        release_object( change );
    }
}
